package auth

import (
	"context"
	"database/sql"
	"errors"
	"fmt"
	"net/http"
	"strings"
	"time"

	cache "github.com/Code-Hex/go-generics-cache"
	"github.com/Code-Hex/go-generics-cache/policy/lru"
	"github.com/Masterminds/squirrel"
	openapi_errors "github.com/go-openapi/errors"
	"github.com/jmoiron/sqlx"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/rs/zerolog"
	"golang.org/x/crypto/bcrypt"
	"gopkg.in/dgrijalva/jwt-go.v3"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/utils"
)

var (
	// ErrInvalidCredentials is returned when auth fails because the credentials are
	// invalid.
	ErrInvalidCredentials = errors.New("invalid credentials")

	// ErrAccountidCobrandidIncompatible is returned when auth fails because the
	// credentials contains both a accountid and cobrandid. This error will
	// disappear in a later version.
	ErrAccountidCobrandidIncompatible = errors.New(
		"credentials cannot contain both accountid and cobrandid")

	// CookieName is the name of the auth cookie.
	CookieName = "rednerd-auth"
)

// TokenOptions holds options related to the JWT.
type TokenOptions struct {
	Secret             string             `long:"token-secret" ini-name:"token-secret" env:"AUTH_TOKEN_SECRET" description:"HMAC secret for signing/verifying JWT"`
	ExpirationOpt      func(string) error `long:"token-expiration" ini-name:"token-expiration" required:"false" description:"Expiration time of the generated tokens" default:"5m"`
	Expiration         time.Duration      `no-flag:"t"`
	CookieDomain       string             `long:"cookie-domain" ini-name:"cookie-domain" description:"The domain used in cookies, must match the public-facing server host name"`
	CookieSecure       bool               `long:"cookie-secure" ini-name:"cookie-secure" description:"Set to true if the cookie must be over https only (recommended)"`
	CacheMaxSize       int                `long:"token-cache-max-size" ini-name:"token-cache-max-size" required:"faise" description:"Maximum number of entries in the token cache" default:"10000"`
	CachePurgeDelay    time.Duration      `no-flag:"t"`
	CachePurgeDelayOpt func(string) error `long:"token-cache-purge-delay" ini-name:"token-cache-purge-delay" required:"false" description:"Delay between token cache purges" default:"10m"`
}

func durationOption(tgt *time.Duration) func(string) error {
	return func(duration string) error {
		d, err := time.ParseDuration(duration)
		if err != nil {
			return err
		}

		*tgt = d

		return nil
	}
}

// NewTokenOptions creates a TokenOptions.
func NewTokenOptions() *TokenOptions {
	options := TokenOptions{}
	options.ExpirationOpt = durationOption(&options.Expiration)
	options.CachePurgeDelayOpt = durationOption(&options.CachePurgeDelay)

	return &options
}

// Auth handles token authentication & creation.
type Auth struct {
	ctx          context.Context
	db           *sqlx.DB
	secret       []byte
	cacheMaxSize int
	expiration   time.Duration
	cookieDomain string
	cookieSecure bool

	parser jwt.Parser

	log zerolog.Logger

	cache              *cache.Cache[string, *models.Principal]
	collectors         []prometheus.Collector
	cacheHitCollector  prometheus.Counter
	cacheMissCollector prometheus.Counter
}

func (auth *Auth) Expiration() time.Duration {
	return auth.expiration
}

// NewAuth creates a Auth instance.
func NewAuth(ctx context.Context, options TokenOptions, db *sqlx.DB, log zerolog.Logger) *Auth {
	if options.CacheMaxSize == 0 {
		options.CacheMaxSize = 100
	}

	if options.Expiration == 0 {
		options.Expiration = time.Minute * 5
	}

	if options.CachePurgeDelay == 0 {
		options.CachePurgeDelay = time.Minute * 10
	}

	if options.Secret == "" {
		log.Warn().Msg(
			`token-secret is not provided. A random one will be used,
            meaning emitted tokens will be invalidated by a restart.`)

		options.Secret = utils.RandomString(32)
	}

	cache := cache.NewContext(
		ctx,
		cache.AsLRU[string, *models.Principal](
			lru.WithCapacity(options.CacheMaxSize),
		),
		cache.WithJanitorInterval[string, *models.Principal](options.CachePurgeDelay),
	)
	auth := Auth{
		ctx,
		db,
		[]byte(options.Secret),
		options.CacheMaxSize,
		options.Expiration,
		options.CookieDomain,
		options.CookieSecure,
		jwt.Parser{ValidMethods: []string{"HS256"}},
		log,
		cache,
		nil,
		prometheus.NewCounter(
			prometheus.CounterOpts{
				Name: "rednerd_auth_token_cache_hit_total",
				Help: "The token cache hits",
			},
		),
		prometheus.NewCounter(
			prometheus.CounterOpts{
				Name: "rednerd_auth_token_cache_miss_total",
				Help: "The token cache misses",
			},
		),
	}

	auth.collectors = []prometheus.Collector{
		prometheus.NewGaugeFunc(
			prometheus.GaugeOpts{
				Name: "rednerd_auth_token_cache_size",
				Help: "The number of entries in the token cache",
			}, func() float64 {
				return float64(auth.cache.Len())
			},
		),
		prometheus.NewGaugeFunc(
			prometheus.GaugeOpts{
				Name: "rednerd_auth_token_cache_size_max",
				Help: "The maximum number of entries in the token cache",
			}, func() float64 {
				return float64(auth.cacheMaxSize)
			},
		),
		auth.cacheHitCollector,
		auth.cacheMissCollector,
	}

	for _, c := range auth.collectors {
		prometheus.MustRegister(c)
	}

	return &auth
}

// Close release the underlying resources.
func (auth *Auth) Close() {
	// unregister prometheus collectors
	for _, c := range auth.collectors {
		prometheus.Unregister(c)
	}
	// empty cache
	for _, key := range auth.cache.Keys() {
		auth.cache.Delete(key)
	}
}

func (auth *Auth) addToCache(token string, principal *models.Principal) {
	auth.cache.Set(
		token,
		principal,
		cache.WithExpiration(time.Until(time.Unix(principal.StandardClaims.ExpiresAt, 0))),
	)
}

// Authenticate authenticates a user.
func (auth *Auth) Authenticate(
	credentials *models.Credentials,
) (*models.Principal, error) {
	log := zerolog.Ctx(auth.ctx)

	if credentials == nil {
		return nil, ErrInvalidCredentials
	}

	var user models.UserDB

	sqlH := database.NewSQLHelper(auth.ctx, auth.db, *log)
	if err := sqlH.Get(&user,
		squirrel.
			Select(models.UserDBColumns...).
			From(models.UserDBTable).
			Where(
				squirrel.Eq{models.UserinfoUsernameColumn: credentials.Username}),
	); err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			return nil, ErrInvalidCredentials
		}

		return nil, err
	}

	if err := bcrypt.CompareHashAndPassword([]byte(user.Password), []byte(credentials.Password)); err != nil {
		return nil, ErrInvalidCredentials
	}

	return &models.Principal{
		StandardClaims: jwt.StandardClaims{
			Subject: user.Username,
		},
		IsAdmin:               user.IsAdmin,
		EnabledRequestLogging: user.EnableRequestLogging,
	}, nil
}

// AuthFromAuthorization loads a 'Authorization' header and returns a Principal instance.
func (auth *Auth) AuthFromAuthorization(header string) (*models.Principal, error) {
	if !strings.HasPrefix(header, "Bearer ") {
		return nil, openapi_errors.New(
			http.StatusUnauthorized,
			"Invalid Authorization header value. Expects 'Bearer <token>', got '%s'",
			header)
	}

	tokenString := header[7:]

	return auth.AuthFromToken(tokenString)
}

// AuthFromCookie loads a 'Cookie' header and returns a Principal instance.
func (auth *Auth) AuthFromCookie(header string) (*models.Principal, error) {
	h := http.Header{}
	h.Add("Cookie", header)
	request := http.Request{Header: h}

	cookie, err := request.Cookie(CookieName)
	if err != nil {
		if errors.Is(err, http.ErrNoCookie) {
			return nil, fmt.Errorf("no %s cookie found", CookieName)
		}

		return nil, err
	}

	return auth.AuthFromToken(cookie.Value)
}

// AuthFromAPIKey loads a 'Rednerd-API-Key' header and returns a Principal instance.
func (auth *Auth) AuthFromAPIKey(header string) (*models.Principal, error) {
	return auth.AuthFromAPIKeyContext(auth.ctx, header)
}

func (auth *Auth) AuthFromAPIKeyContext(ctx context.Context, header string) (*models.Principal, error) {
	userColumns := make([]string, len(models.UserDBColumns))
	for i, c := range models.UserDBColumns {
		userColumns[i] = models.UserDBTable + "." + c
	}

	var user models.UserDB

	sqlH := database.NewSQLHelper(ctx, auth.db, auth.log)
	if err := sqlH.Get(&user,
		squirrel.
			Select(userColumns...).
			From(models.UserDBTable).
			LeftJoin(database.Join(
				models.UserDBTable,
				models.UserDBUsernameColumn,
				models.ApikeyDBTable,
				models.ApikeyDBUsernameColumn)).
			Where(
				squirrel.Eq{models.ApikeyDBValueColumn: header}),
	); err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			return nil, ErrInvalidCredentials
		}

		return nil, err
	}

	principal := &models.Principal{
		StandardClaims: jwt.StandardClaims{
			Subject: user.Username,
		},
		IsAdmin:               user.IsAdmin,
		EnabledRequestLogging: user.EnableRequestLogging,
	}

	return principal, nil
}

// AuthFromToken loads a token.
func (auth *Auth) AuthFromToken(tokenString string) (*models.Principal, error) {
	if p, ok := auth.cache.Get(tokenString); ok {
		auth.cacheHitCollector.Inc()

		return p, nil
	}

	auth.cacheMissCollector.Inc()

	token, err := auth.parser.ParseWithClaims(
		tokenString,
		&models.Principal{},
		func(token *jwt.Token) (interface{}, error) {
			return auth.secret, nil
		})
	if err != nil {
		return nil, openapi_errors.New(http.StatusUnauthorized, "Invalid token: %s", err)
	}

	if !token.Valid {
		return nil, openapi_errors.New(http.StatusUnauthorized, "Invalid token")
	}

	claims, ok := token.Claims.(*models.Principal)
	if !ok {
		panic("could not get principal from claims")
	}

	auth.addToCache(tokenString, claims)

	return claims, nil
}

// MakeToken creates a new token.
func (auth *Auth) MakeToken(principal models.Principal, withCookie bool) (string, *http.Cookie, error) {
	expiresIn := auth.expiration
	if withCookie {
		expiresIn *= 10
	}

	expiresAt := time.Now().Add(expiresIn)
	principal.StandardClaims.ExpiresAt = expiresAt.Unix()
	token := jwt.NewWithClaims(jwt.SigningMethodHS256, principal)

	stoken, err := token.SignedString(auth.secret)
	if err != nil {
		return "", nil, err
	}

	auth.addToCache(stoken, &principal)

	var cookie *http.Cookie
	if withCookie {
		cookie = &http.Cookie{
			Name:     CookieName,
			Value:    stoken,
			Expires:  expiresAt,
			Path:     "/",
			Domain:   auth.cookieDomain,
			Secure:   auth.cookieSecure,
			HttpOnly: true,
		}
	}

	return stoken, cookie, nil
}

func (auth *Auth) Logout() *http.Cookie {
	cookie := http.Cookie{
		Name:     CookieName,
		MaxAge:   -1,
		Path:     "/",
		Domain:   auth.cookieDomain,
		Secure:   auth.cookieSecure,
		HttpOnly: true,
	}

	return &cookie
}
