package restapi

import (
	"net/http"
	"strings"
	"time"

	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"

	"orus.io/orus-io/rednerd/auth"
	"orus.io/orus-io/rednerd/models"
)

var publicAccess = []string{
	"/api/v1/auth/authenticate",
	"/api/v1/auth/register",
	"/api/v1/auth/reset-password",
	"/api/v1/auth/forgot-password",
	"/api/docs",
	"/api/docs/",
	"/swagger.json",
	"/metrics",
}

func matchPublicAccess(path string) bool {
	for _, p := range publicAccess {
		if p == path || strings.HasSuffix(p, "/") && strings.HasPrefix(path, p) {
			return true
		}
	}

	return false
}

// AuthMiddleware authenticates the user.
func AuthMiddleware(auth *auth.Auth, db *sqlx.DB, log zerolog.Logger) func(http.Handler) http.Handler {
	return func(next http.Handler) http.Handler {
		return http.HandlerFunc(func(rw http.ResponseWriter, r *http.Request) {
			log := zerolog.Ctx(r.Context())
			authHeader := r.Header.Get("Authorization")
			apikeyHeader := r.Header.Get("Rednerd-API-Key")

			var principal *models.Principal

			if cookie, err := r.Cookie("rednerd-auth"); err == nil {
				pr, err := auth.AuthFromToken(cookie.Value)
				if err != nil {
					log.Err(err).Msg("failed to read the rednerd-auth cookie")
				} else {
					principal = pr
					if time.Until(time.Unix(principal.StandardClaims.ExpiresAt, 0)) < auth.Expiration()*4/5 {
						_, cookie, err := auth.MakeToken(*principal, true)
						if err != nil {
							log.Err(err).Msg("failed to create the rednerd-auth cookie")
						} else {
							http.SetCookie(rw, cookie)
						}
					}
				}
			}

			switch {
			case authHeader != "":
				pr, err := auth.AuthFromAuthorization(authHeader)
				if err != nil {
					log.Err(err).Msg("error reading authorization header")
				} else {
					principal = pr
				}
			case apikeyHeader != "":
				pr, err := auth.AuthFromAPIKeyContext(r.Context(), apikeyHeader)
				if err != nil {
					log.Err(err).Msg("error reading api key header")
				} else {
					principal = pr
				}
			default:
				if matchPublicAccess(r.URL.Path) {
					next.ServeHTTP(rw, r)

					return
				}
			}

			if principal != nil {
				log.UpdateContext(func(c zerolog.Context) zerolog.Context {
					return c.Object("user", principal)
				})
				next.ServeHTTP(rw, r)

				return
			}

			rw.WriteHeader(http.StatusUnauthorized)
			_, _ = rw.Write([]byte("Unauthorized"))
		})
	}
}
