package handlers

import (
	"context"
	"fmt"

	"github.com/Masterminds/squirrel"
	"github.com/go-openapi/runtime/middleware"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/restapi/operations/template_management"
)

// NewAccountTemplateListHandler returns a AccountTemplateListHandler.
func NewAccountTemplateListHandler(db *sqlx.DB) *AccountTemplateListHandler {
	return &AccountTemplateListHandler{
		db: db,
	}
}

// AccountTemplateListHandler handles GET /user.
type AccountTemplateListHandler struct {
	db *sqlx.DB
}

func (h *AccountTemplateListHandler) handle(
	ctx context.Context, account string,
) ([]*models.Template, error) {
	log := zerolog.Ctx(ctx)

	var templatedbList []models.TemplateDB

	sql := database.NewSQLHelper(ctx, h.db, *log)

	if err := sql.Select(&templatedbList,
		squirrel.Select(templateColumnsWithoutBody...).
			From(models.TemplateDBTable).
			Where(
				squirrel.Eq{
					models.TemplateDBAccountColumn: account,
				}),
	); err != nil {
		return nil, err
	}

	list := make([]*models.Template, 0, len(templatedbList))

	for i := range templatedbList {
		if err := HandleTemplateActivity(ctx, h.db, &templatedbList[i], log); err != nil {
			return nil, err
		}

		list = append(list, &templatedbList[i].Template)
	}

	return list, nil
}

// Handle handles the incoming request.
func (h *AccountTemplateListHandler) Handle(
	params template_management.AccountTemplateListParams, principal *models.Principal,
) middleware.Responder {
	ctx := params.HTTPRequest.Context()

	log := zerolog.Ctx(ctx)
	if principal.IsAdmin {
		log.Error().Msg("admin cannot manage templates. Please use a non-admin user")

		return template_management.
			NewAccountTemplateAddUnauthorized()
	}

	if principal.Subject != params.Account {
		log.Error().Msg("user can manage its own templates only")

		return template_management.
			NewAccountTemplateAddUnauthorized()
	}

	templates, err := h.handle(ctx, params.Account)

	if err != nil {
		return InternalError(err, log, false)
	}

	return template_management.NewAccountTemplateListOK().WithPayload(templates)
}

var templateColumnsWithoutBody = make([]string, len(models.TemplateDBColumns))

func init() {
	var (
		columns  = models.TemplateDBColumns
		colcount = len(columns)
	)

	if columns[colcount-1] != models.TemplateDBBytesBodyColumn {
		panic(fmt.Sprintf("%s is not %s", columns[colcount-1], models.TemplateDBBytesBodyColumn))
	}

	templateColumnsWithoutBody = columns[:colcount-1]
}
