package handlers

import (
	"context"
	"errors"
	"fmt"

	"github.com/Masterminds/squirrel"
	"github.com/go-openapi/runtime/middleware"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"orus.io/orus-io/go-orusapi/database"

	redner "orus.io/orus-io/rednerd/lib"
	"orus.io/orus-io/rednerd/models"
	op "orus.io/orus-io/rednerd/restapi/operations/rendering"
	"orus.io/orus-io/rednerd/utils"
)

var (
	// ErrNoTemplateNorDocuments is returned if the render request has no template
	// nor a document.
	ErrNoTemplateNorDocument = errors.New(
		"no template nor document. Must have one",
	)
	// ErrBothTemplateAndDocument is returned if the render request has a template
	// and a document.
	ErrBothTemplateAndDocument = errors.New(
		"render_request cannot accept both a Document and a Template",
	)
)

// NewRenderHandler creates a RenderHandler.
func NewRenderHandler(renderer *redner.Renderer, db *sqlx.DB) *RenderHandler {
	return &RenderHandler{renderer, db}
}

// RenderHandler handles POST /api/v1/render.
type RenderHandler struct {
	renderer *redner.Renderer
	db       *sqlx.DB
}

func (h *RenderHandler) handle(
	ctx context.Context,
	log zerolog.Logger,
	principal *models.Principal,
	template *models.Template,
	data models.Dataset,
	document *models.Document,
	metadata models.Metadata,
	toType string,
) ([]*models.Document, error) {
	switch {
	case template == nil && document == nil:
		return nil, utils.HTTPBadRequest(ErrNoTemplateNorDocument)
	case template != nil && document != nil:
		return nil, utils.HTTPBadRequest(ErrBothTemplateAndDocument)
	default:
	}

	if template != nil && template.Name != "" {
		var tDB models.TemplateDB
		db := database.NewSQLHelper(ctx, h.db, log)
		q := squirrel.Select(models.TemplateDBColumns...).
			From(models.TemplateDBTable).
			Where(squirrel.Eq{models.TemplateDBNameColumn: template.Name})
		if err := db.Get(&tDB, q); err != nil {
			return nil, err
		}
		if tDB.Account != principal.Subject {
			return nil, errors.New("user can render on it's own templates only")
		}
	}

	documents, err := h.renderer.Render(
		ctx, principal.Subject, template, data, document, metadata, toType,
		principal.EnabledRequestLogging,
	)
	if errors.Is(err, redner.ErrInvalidInput) {
		return nil, utils.HTTPBadRequest(err)
	} else if err != nil {
		return nil, err
	}

	return documents, nil
}

// Handle the incoming request.
func (h *RenderHandler) Handle(params op.RenderParams, principal *models.Principal) middleware.Responder {
	req := params.Request
	accept := req.Accept
	if accept == "" {
		acceptHeader := params.HTTPRequest.Header.Get("Accept")
		fmt.Printf("hey %s", acceptHeader)
		accept = models.Accept(acceptHeader)
	}
	ctx := params.HTTPRequest.Context()
	log := zerolog.Ctx(ctx)
	docs, err := h.handle(
		ctx,
		*log,
		principal,
		req.Template,
		req.Data,
		req.Document,
		req.Metadata,
		string(accept),
	)
	if err != nil {
		var httpErr *utils.HTTPError
		if errors.As(err, &httpErr) {
			r := op.NewRenderDefault(httpErr.StatusCode)
			if httpErr.Err != nil {
				r.WithPayload(models.FromError(httpErr.Err))
			}

			return r
		}

		if ctx.Err() != nil {
			return op.NewRenderDefault(499)
		}

		return InternalError(err, log, false)
	}
	fmt.Print("heyyyyy")

	return op.NewRenderOK().WithPayload(docs)
}
