package restapi

import (
	"bytes"
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"net/http"
	"os"
	"runtime/debug"

	"github.com/getsentry/sentry-go"
	jsoniter "github.com/json-iterator/go"
	"github.com/rs/zerolog"
)

// CatchPanics catches the panics and log them as errors.
func CatchPanics(next http.Handler) http.Handler {
	return http.HandlerFunc(func(rw http.ResponseWriter, req *http.Request) {
		ctx := req.Context()
		defer func(ctx context.Context) {
			if r := recover(); r != nil {
				log := zerolog.Ctx(ctx)

				if err, ok := r.(error); ok {
					if errors.Is(err, os.ErrDeadlineExceeded) {
						log.Warn().Err(err).Msg("Deadline exceeded, handler probably too slow")

						return
					}
				}

				exc := zerolog.Dict().
					Str("value", fmt.Sprintf("%#v", r)).
					Str("type", fmt.Sprintf("%T", r))

				stack := sentry.NewStacktrace()
				stack.Frames = stack.Frames[:len(stack.Frames)-1]

				b, err := jsoniter.Marshal(stack)
				if err != nil {
					log.Err(err).Msg("Could not marshal stacktrace, fallback to plain debug.Stack()")

					b = debug.Stack()
					exc = exc.Str("plainstack", string(b))
				} else {
					exc = exc.RawJSON("stacktrace", b)
					var prettyJSON bytes.Buffer
					err := json.Indent(&prettyJSON, b, "", "\t")
					if err != nil {
						log.Err(err).Msg("")
					}
					log.Info().Msgf("pretty %s", prettyJSON.String())
				}

				log.Error().Dict("exception", exc).Msg("Caught a panic")

				rw.WriteHeader(http.StatusInternalServerError)
			}
		}(ctx)
		next.ServeHTTP(rw, req)
	})
}
